#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : pdf
# @Time         : 2023/6/30 18:58
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from langchain.document_loaders.parsers.pdf import BaseBlobParser, Blob, Document
from langchain.document_loaders.pdf import PyMuPDFLoader as _PyMuPDFLoader


class PyMuPDFLoader(_PyMuPDFLoader):
    """Loader that uses PyMuPDF to load PDF files."""

    def load(self, **kwargs: Optional[Any]) -> List[Document]:
        """Load file."""

        parser = PyMuPDFParser(text_kwargs=kwargs, page_preprocessing=self.page_preprocessing)

        blob = Blob.from_path(self.file_path)
        return parser.parse(blob)

    @staticmethod
    @lru_cache(1024)
    def page_preprocessing(page_content, re_parser=None):  # todo: 丰富预处理逻辑
        page_content = page_content.strip().lower()
        if re_parser:
            page_content = re_parser.sub(page_content, '')

        return page_content


class PyMuPDFParser(BaseBlobParser):
    """Parse PDFs with PyMuPDF."""

    def __init__(self, text_kwargs: Optional[Mapping[str, Any]] = None, page_preprocessing=None) -> None:
        """Initialize the parser.

        Args:
            text_kwargs: Keyword arguments to pass to ``fitz.Page.get_text()``.
        """
        self.text_kwargs = text_kwargs or {}
        self.page_preprocessing = page_preprocessing

    def lazy_parse(self, blob: Blob) -> Iterator[Document]:
        """Lazily parse the blob."""
        import fitz

        with blob.as_bytes_io() as file_path:
            doc = fitz.open(file_path)  # open document

            for page in doc:
                page_content = page.get_text(**self.text_kwargs).strip()
                page_content = self.page_preprocessing(page_content)  # 文本预处理
                if not page_content:
                    continue

                yield Document(
                    page_content=page_content,
                    metadata={
                        **{
                            "source": blob.source,
                            "file_path": blob.source,
                            "page": page.number,
                            "total_pages": len(doc),

                            "page_length": len(page_content),  #

                        },
                        **{
                            k: doc.metadata[k]
                            for k in doc.metadata
                            if type(doc.metadata[k]) in [str, int]
                        },
                    },
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : __init__.py
# @Time         : 2023/6/30 16:37
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from meutils.cache_utils import diskcache


def init(verbose=-1):
    import openai

    OPENAI_CACHE = os.getenv("OPENAI_CACHE", "~/.cache/openai_cache")

    openai.Embedding.create = diskcache(
        openai.Embedding.create,
        location=f"{OPENAI_CACHE}_EMBEDDINGS",
        verbose=verbose,
    )

    # 流式会生成不了
    # openai.Completion.create = diskcache(
    #     openai.Completion.create,
    #     location=f"{OPENAI_CACHE}_Completion",
    #     verbose=verbose,
    #     ttl=24 * 3600
    # )
    #
    # openai.ChatCompletion.create = diskcache(
    #     openai.ChatCompletion.create,
    #     location=f"{OPENAI_CACHE}_ChatCompletion",
    #     verbose=verbose,
    #     ttl=24 * 3600
    # )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : datamodels
# @Time         : 2023/6/30 18:23
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import numpy as np

from meutils.pipe import *


class Metadata(BaseModel):
    source: str
    file_name: str
    category: str
    subcategory: str
    text: str

    keywords: List[str]  # 根据查询功能适配数据类型

    part_id: int
    part_length: int
    total_length: int

    ext: dict = Field(default_factory=dict)  # $meta



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : 缓存
# @Time         : 2023/7/4 10:59
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : https://python.langchain.com/docs/modules/model_io/models/llms/integrations/llm_caching#gptcache
from langchain.schema import HumanMessage

from meutils.pipe import *
import time


def response_text(openai_resp):
    return openai_resp['choices'][0]['message']['content']


print("Cache loading.....")

# To use GPTCache, that's all you need
# -------------------------------------------------
from gptcache import cache
from gptcache.adapter import openai

cache.init()
# -------------------------------------------------

question = "周杰伦是谁，50字总结"
for _ in range(2):
    start_time = time.time()
    response = openai.ChatCompletion.create(
        model='gpt-3.5-turbo',
        messages=[
            {
                'role': 'user',
                'content': question
            }
        ],
    )
    print(f'Question: {question}')
    print("Time consuming: {:.2f}s".format(time.time() - start_time))
    print(f'Answer: {response_text(response)}\n')



from gptcache import cache
from gptcache.processor.pre import get_prompt
# init gptcache
cache.init(pre_embedding_func=get_prompt)

from langchain.llms import OpenAI
from gptcache.adapter.langchain_models import LangChainLLMs
# run llm with gptcache
llm = LangChainLLMs(llm=OpenAI(temperature=0))
llm("Hello world")


from gptcache import cache
from gptcache.processor.pre import get_messages_last_content
# init gptcache
cache.init(pre_embedding_func=get_messages_last_content)
cache.set_openai_key()
from langchain.chat_models import ChatOpenAI
from gptcache.adapter.langchain_models import LangChainChat
# run chat model with gptcache
chat = LangChainChat(chat=ChatOpenAI(temperature=0))
chat([HumanMessage(content="Translate this sentence from English to French. I love programming.")])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : baidu
# @Time         : 2023/7/24 13:22
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
from meutils.pipe import *

from typing import Mapping, Any, List, Optional, Tuple, Dict
import requests
import json
import os
import sys

from langchain.chat_models.base import BaseChatModel
from langchain.schema import BaseMessage, ChatResult, HumanMessage, AIMessage, SystemMessage, ChatMessage, ChatGeneration

sys.path.append(os.path.join(os.path.dirname(__file__), '../..'))


CHAT_CONFIG = CHAT_CONFIG['ernie']
llm_kwargs = CHAT_CONFIG.get('llm_kwargs', {})


class ChatLLM(BaseChatModel):
    '''Chat with LLM given context. Must be a LangChain BaseLanguageModel to adapt agent.'''
    api_key: str = CHAT_CONFIG['ernie_api_key']
    secret_key: str = CHAT_CONFIG['ernie_secret_key']
    temperature: float = llm_kwargs.get('temperature', 0)
    max_tokens: Optional[int] = llm_kwargs.get('max_tokens', None)
    n: int = llm_kwargs.get('n', 1)

    def _generate(self, messages: List[BaseMessage], stop: Optional[List[str]] = None) -> ChatResult:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params['messages'] = message_dicts
        payload = json.dumps(params)
        headers = {
            'Content-Type': 'application/json'
        }

        url = self._create_url()
        response = requests.request(
            'POST', url, headers=headers, data=payload)
        return self._create_chat_result(response)

    async def _agenerate(self, messages: List[BaseMessage], stop: Optional[List[str]] = None) -> ChatResult:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params['messages'] = message_dicts
        payload = json.dumps(params)
        headers = {
            'Content-Type': 'application/json'
        }
        url = self._create_url()
        response = requests.request(
            'POST', url, headers=headers, data=payload)
        return self._create_chat_result(response)

    def _create_url(self):
        access_token = self._get_access_token(
            api_key=self.api_key, secret_key=self.secret_key)
        url = 'https://aip.baidubce.com/rpc/2.0/ai_custom/v1/wenxinworkshop/chat/completions?access_token=' \
            + access_token
        return url

    def _create_message_dicts(
        self, messages: List[BaseMessage], stop: Optional[List[str]]
    ) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        params: Dict[str, Any] = {**self._default_params}
        if stop is not None:
            if 'stop' in params:
                raise ValueError(
                    '`stop` found in both the input and default params.')
            params['stop'] = stop
        message_dicts = []
        for m in messages:
            message_dicts.append(self._convert_message_to_dict(m))
            if isinstance(m, SystemMessage):
                message_dicts.append(
                    {'role': 'assistant', 'content': 'OK.'}
                )
        return message_dicts, params

    def _create_chat_result(self, response: Mapping[str, Any]) -> ChatResult:
        generations = []
        response = response.json()
        if 'result' not in response:
            raise RuntimeError(response)
        message = self._convert_dict_to_message(
            {'role': 'assistant', 'content': response['result']})
        gen = ChatGeneration(message=message)
        generations.append(gen)
        llm_output = {
            'token_usage': response['usage'], 'model_name': 'ernie'}
        return ChatResult(generations=generations, llm_output=llm_output)

    @staticmethod
    def _convert_message_to_dict(message: BaseMessage) -> dict:
        if isinstance(message, ChatMessage):
            message_dict = {'role': message.role, 'content': message.content}
        elif isinstance(message, (HumanMessage, SystemMessage)):
            message_dict = {'role': 'user', 'content': message.content}
        elif isinstance(message, AIMessage):
            message_dict = {'role': 'assistant', 'content': message.content}
        else:
            raise ValueError(f'Got unknown type {message}')
        if 'name' in message.additional_kwargs:
            message_dict['name'] = message.additional_kwargs['name']
        return message_dict

    @staticmethod
    def _convert_dict_to_message(_dict: dict) -> BaseMessage:  # pylint: disable=C0103
        role = _dict['role']
        if role == 'user':
            return HumanMessage(content=_dict['content'])
        elif role == 'assistant':
            return AIMessage(content=_dict['content'])
        elif role == 'system':
            return SystemMessage(content=_dict['content'])
        else:
            return ChatMessage(content=_dict['content'], role=role)

    @staticmethod
    def _get_access_token(api_key, secret_key):
        url = 'https://aip.baidubce.com/oauth/2.0/token'
        params = {
            'grant_type': 'client_credentials',
            'client_id': api_key,
            'client_secret': secret_key
        }
        return str(requests.post(url, params=params).json().get('access_token'))

    @property
    def _default_params(self) -> Dict[str, Any]:
        '''Get the default parameters for calling OpenAI API.'''
        return {
            'max_tokens': self.max_tokens,
            'n': self.n,
            'temperature': self.temperature,
        }

    @property
    def _llm_type(self) -> str:
        return 'ernie'


# if __name__ == '__main__':
#     chat = ChatLLM()
#     messages = [HumanMessage(content='Translate this sentence from English to French. I love programming.')]
#     ans = chat(messages)
#     print(type(ans), ans)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : llmcache
# @Time         : 2023/7/6 11:51
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : https://zhuanlan.zhihu.com/p/629348339

from meutils.pipe import *

from gptcache import cache
from gptcache.embedding import Onnx, OpenAI
from gptcache.manager import CacheBase, VectorBase, get_data_manager
from gptcache.similarity_evaluation import SearchDistanceEvaluation, ExactMatchEvaluation
from gptcache.processor.pre import get_prompt
from gptcache.adapter.api import init_similar_cache


def get_msg_func(data, **_):
    return data.get("messages")[-1].content


onnx = Onnx()
cache_base = CacheBase('sqlite')
vector_base = VectorBase('faiss', dimension=onnx.dimension)
data_manager = get_data_manager(cache_base, vector_base)
cache.init(
    pre_embedding_func=get_msg_func,
    embedding_func=onnx.to_embeddings,
    data_manager=data_manager,
    similarity_evaluation=SearchDistanceEvaluation(),
)
cache.set_openai_key()


from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.llms import OpenAI



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : common
# @Time         : 2023/7/4 08:56
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from langchain import LLMChain, OpenAI, PromptTemplate
from langchain.document_loaders.base import Document

from langchain.callbacks import AsyncIteratorCallbackHandler
from langchain.chains.base import Chain

# prompt_template = "Tell me a {adjective} joke"
# prompt = PromptTemplate(
#     input_variables=["adjective"], template=prompt_template
# )

template2prompt = PromptTemplate.from_template


def docs2dataframe(docs: List[Document]) -> pd.DataFrame:
    return pd.DataFrame(map(lambda doc: {**doc.metadata, **{'page_content': doc.page_content}}, docs))


def dataframe2docs(df: pd.DataFrame) -> List[Document]:
    df = df.copy()
    docs = []
    for page_content, metadata in zip(df.pop('page_content'), df.to_dict(orient='records')):
        docs.append(Document(page_content=page_content, metadata=metadata))
    return docs


def get_api_key(n: int = 1) -> List[str]:
    """获取keys"""
    openai_api_key_set = (
        os.getenv("OPENAI_API_KEY_SET", "").replace(' ', '').strip(',').strip().split(',') | xfilter | xset
    )
    openai_api_key_path = os.getenv("OPENAI_API_KEY_PATH", '')
    if Path(openai_api_key_path).is_file():
        openai_api_key_set = set(Path(openai_api_key_path).read_text().strip().split())
    return list(openai_api_key_set)[:n]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : Embeddings
# @Time         : 2023/6/30 17:13
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from langchain.embeddings import OpenAIEmbeddings, HuggingFaceEmbeddings



from langchain.chains import ChatVectorDBChain, VectorDBQAWithSourcesChain, ConversationalRetrievalChain, ConversationChain





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : demo
# @Time         : 2023/6/30 17:26
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.text_splitter import CharacterTextSplitter
from langchain.vectorstores import Milvus
from langchain.indexes import VectorstoreIndexCreator
from langchain.vectorstores import Chroma, Milvus, DocArrayInMemorySearch, FAISS
from langchain.embeddings import OpenAIEmbeddings
from langchain import OpenAI
from langchain.llms import OpenAI, OpenAIChat
from langchain.chat_models import ChatOpenAI
from langchain.indexes import VectorstoreIndexCreator
from langchain.chains import RetrievalQA, RetrievalQAWithSourcesChain

from langchain.chains.question_answering import load_qa_chain



from llama_index import (
    SimpleDirectoryReader,
    LLMPredictor,
    ServiceContext,
    ResponseSynthesizer
)
from llama_index.indices.document_summary import GPTDocumentSummaryIndex
from langchain.chat_models import ChatOpenAI
from llama_index.indices.keyword_table import KeywordTableIndex, KeywordTableSimpleRetriever, KeywordTableRAKERetriever, \
    SimpleKeywordTableIndex as _SimpleKeywordTableIndex

def extract_keywords(text, max_keywords):
    return {}


class SimpleKeywordTableIndex(_SimpleKeywordTableIndex):
    def _extract_keywords(self, text: str) -> Set[str]:
        """Extract keywords from text."""
        return extract_keywords(text, self.max_keywords_per_chunk)


from llama_index import (
    VectorStoreIndex,
    # GPTVectorStoreIndex,
    SimpleDirectoryReader,
    ServiceContext,
    LLMPredictor,
    PromptHelper,
    Document,
    StorageContext,
    load_index_from_storage, MockLLMPredictor, GPTTreeIndex, MockEmbedding, ResponseSynthesizer, Prompt
)
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler


StreamingStdOutCallbackHandler


from langchain.chains import RetrievalQA


from langchain import OpenAI
from langchain.llms import OpenAI


from llama_index import (
    VectorStoreIndex,
    # GPTVectorStoreIndex,
    SimpleDirectoryReader,
    ServiceContext,
    LLMPredictor,
    PromptHelper,
    Document,
    StorageContext,
    load_index_from_storage, MockLLMPredictor, GPTTreeIndex, MockEmbedding, ResponseSynthesizer, Prompt
)


import paddlenlp.taskflow


from langchain.llms import OpenAI, OpenAIChat
from langchain import OpenAI
from langchain.chat_models import ChatOpenAI



from langchain import OpenAI
from llama_index import VectorStoreIndex, SimpleDirectoryReader
from langchain.document_loaders import DirectoryLoader, UnstructuredFileLoader, PyMuPDFLoader

import fastchat



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : ChatOpenAI
# @Time         : 2023/7/4 13:54
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  : 异步流式输出 https://github.com/sugarforever/LangChain-Tutorials/blob/main/StreamChat/app.py

# from langchain.callbacks import AsyncIteratorCallbackHandler
# callback = AsyncIteratorCallbackHandler()
# model = ChatOpenAI(streaming=True, verbose=True, callbacks=[callback])
# coroutine = model.agenerate(messages=[[HumanMessage(content='1+111')]])
# task = asyncio.create_task(coroutine)
# async for token in callback.aiter():
#     print(f"{token}", end='')

from meutils.pipe import *
from langchain.chat_models import ChatOpenAI as _ChatOpenAI
from langchain.callbacks.streaming_aiter import AsyncIteratorCallbackHandler

from langchain.chains import ConversationalRetrievalChain, VectorDBQAWithSourcesChain, RetrievalQAWithSourcesChain, RetrievalQA, MultiRetrievalQAChain



from langchain.memory import ConversationBufferMemory
from langchain.chains.conversation.memory import ConversationSummaryMemory


#
# qa = RetrievalQA.from_chain_type(
#     llm=llm, chain_type="stuff",
#     retriever=retriever1,
#     return_source_documents=True, memory=memory
# )
